@@ -59,7 +59,7 @@ static const GOptionEntry entries[] = {
 typedef struct _YelpApplicationLoad YelpApplicationLoad;
 struct _YelpApplicationLoad {
     YelpApplication *app;
-    guint timestamp;
+    guint32 timestamp;
     gboolean new;
 };
 
@@ -383,7 +383,7 @@ yelp_application_run (YelpApplication  *app,
 gboolean
 yelp_application_load_uri (YelpApplication  *app,
                            const gchar      *uri,
-                           guint             timestamp,
+                           guint32           timestamp,
                            GError          **error)
 {
     YelpApplicationLoad *data;
@@ -392,6 +392,7 @@ yelp_application_load_uri (YelpApplication  *app,
     data = g_new (YelpApplicationLoad, 1);
     data->app = app;
     data->timestamp = timestamp;
+    data->new = FALSE;
     
     yuri = yelp_uri_new (uri);
     
@@ -407,19 +408,26 @@ void
 yelp_application_new_window (YelpApplication  *app,
                              const gchar      *uri)
 {
-    YelpApplicationLoad *data;
     YelpUri *yuri;
 
-    data = g_new (YelpApplicationLoad, 1);
-    data->app = app;
-    data->new = TRUE;
-
     yuri = yelp_uri_new (uri ? uri : DEFAULT_URI);
 
-    g_signal_connect (yuri, "resolved",
+    yelp_application_new_window_uri (app, yuri);
+}
+
+void
+yelp_application_new_window_uri (YelpApplication  *app,
+                                 YelpUri          *uri)
+{
+    YelpApplicationLoad *data;
+    data = g_new (YelpApplicationLoad, 1);
+    data->app = app;
+    data->timestamp = gtk_get_current_event_time ();
+    data->new = TRUE;
+    g_signal_connect (uri, "resolved",
                       G_CALLBACK (application_uri_resolved),
                       data);
-    yelp_uri_resolve (yuri);
+    yelp_uri_resolve (uri);
 }
 
 static void
@@ -471,7 +479,15 @@ application_uri_resolved (YelpUri             *uri,
     if (gdk_window)
         gdk_x11_window_move_to_current_desktop (gdk_window);
 
-    gtk_window_present_with_time (GTK_WINDOW (window), GDK_CURRENT_TIME);
+    /* Ensure we actually present the window when invoked from the command
+     * line. This is somewhat evil, but the minor evil of Yelp stealing
+     * focus (after you requested it) is outweighed for me by the major
+     * evil of no help window appearing when you click Help.
+     */
+    if (data->timestamp == 0)
+        data->timestamp = gdk_x11_get_server_time (gtk_widget_get_window (GTK_WIDGET (window)));
+
+    gtk_window_present_with_time (GTK_WINDOW (window), data->timestamp);
 
     g_free (data);
 }
@@ -570,7 +586,6 @@ yelp_application_add_bookmark (YelpApplication   *app,
                                const gchar       *title)
 {
     GSettings *settings;
-    YelpApplicationPrivate *priv = GET_PRIV (app);
 
     settings = application_get_doc_settings (app, doc_uri);
 
@@ -596,8 +611,51 @@ yelp_application_add_bookmark (YelpApplication   *app,
             g_variant_builder_add (&builder, "(sss)", page_id, icon, title);
             value = g_variant_builder_end (&builder);
             g_settings_set_value (settings, "bookmarks", value);
-            g_variant_unref (value);
-            g_signal_emit (app, signals[BOOKMARKS_CHANGED], 0, doc_uri, 0);
+            g_signal_emit (app, signals[BOOKMARKS_CHANGED], 0, doc_uri);
+        }
+    }
+}
+
+void
+yelp_application_update_bookmarks (YelpApplication   *app,
+                                   const gchar       *doc_uri,
+                                   const gchar       *page_id,
+                                   const gchar       *icon,
+                                   const gchar       *title)
+{
+    GSettings *settings;
+
+    settings = application_get_doc_settings (app, doc_uri);
+
+    if (settings) {
+        GVariantBuilder builder;
+        GVariantIter *iter;
+        gchar *this_id, *this_icon, *this_title;
+        gboolean updated = FALSE;
+        g_settings_get (settings, "bookmarks", "a(sss)", &iter);
+        g_variant_builder_init (&builder, G_VARIANT_TYPE ("a(sss)"));
+        while (g_variant_iter_loop (iter, "(&s&s&s)", &this_id, &this_icon, &this_title)) {
+            if (g_str_equal (page_id, this_id)) {
+                if (icon && !g_str_equal (icon, this_icon)) {
+                    this_icon = (gchar *) icon;
+                    updated = TRUE;
+                }
+                if (title && !g_str_equal (title, this_title)) {
+                    this_title = (gchar *) title;
+                    updated = TRUE;
+                }
+                if (!updated)
+                    break;
+            }
+            g_variant_builder_add (&builder, "(sss)", this_id, this_icon, this_title);
+        }
+        g_variant_iter_free (iter);
+
+        if (updated) {
+            GVariant *value;
+            value = g_variant_builder_end (&builder);
+            g_settings_set_value (settings, "bookmarks", value);
+            g_signal_emit (app, signals[BOOKMARKS_CHANGED], 0, doc_uri);
         }
     }
 }
@@ -638,7 +696,6 @@ window_resized (YelpWindow        *window,
                 YelpApplication   *app)
 {
     YelpApplicationPrivate *priv = GET_PRIV (app);
-    gint width, height;
     YelpUri *uri;
     gchar *doc_uri;
     GSettings *settings;
